/*
 * This file is part of tela the Tensor Language.
 * Copyright (c) 1994 Pekka Janhunen
 */

#ifdef __GNUC__
#  pragma implementation "objarithm.H"
#endif
#include "objarithm.H"
#include <ctype.h>

static void LengthError(const char *op, const Tobject& a, const Tobject& b) {
	err << "Dimension mismatch with " << Tshort(a) << ' ' << op << ' ' << Tshort(b) << '\n';
	error();
}

static void UnaryError(const char *op, const Tobject& a) {
	err << "Unary '" << op << "' applied on " << Tshort(a) << '\n';
	error();
}

static void BinError(const char *op, const Tobject& a, const Tobject& b) {
	err << "Bad binary operator arguments: " << Tshort(a) << ' ' << op << ' ' << Tshort(b) << '\n';
	error();
}

#define pair(ka,kb) ((ka)*12+(kb))

// Binary operators +,-,*,/,Pow
// '/' is treated using templ/tbinrealop.C, others templ/tbinop.C
// This is because '/' is not integer but real division.

#define OP(a,b) ((a)+(b))
#define OPCHAR "+"
#define OPFNAME Add
#include "templ/tbinop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a,b) ((a)-(b))
#define OPCHAR "-"
#define OPFNAME Sub
#include "templ/tbinop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a,b) ((a)*(b))
#define OPCHAR "*"
#define OPFNAME Mul
#include "templ/tbinop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

// The Div operation does not fit the standard templates
// since the result is always at least real and because
// we want to replace division by inverse multiplication
// whenever possible.

void Div(Tobject& c, const Tobject& a, const Tobject& b) {
	int i,N;
	Treal X;
	Tcomplex Z;
	static char OPCHAR[] = "/";
	switch (pair(a.kind(),b.kind())) {
	case pair(Kint,Kint):
		c = Treal(a.IntValue())/Treal(b.IntValue());
		global::nops++;
		break;
	case pair(Kint,Kreal):
		c = Treal(a.IntValue())/b.RealValue();
		global::nops++;
		break;
	case pair(Kint,Kcomplex):
		c = Tcomplex(a.IntValue())/b.ComplexValue();
		global::nops++;
		break;
	case pair(Kint,KIntArray):
		{
			X = Treal(a.IntValue());
			N = b.length();
			Tint *bp = b.IntPtr();
			if (&b == &c) {
				Tobject c1;
				c1.copydimsRealArray(b);
				Treal *c1p = c1.RealPtr();
				for (i=0; i<N; i++) c1p[i] = X/Treal(bp[i]);
				c = c1;
			} else {
				c.copydimsRealArray(b);
				Treal *cp = c.RealPtr();
				for (i=0; i<N; i++) cp[i] = X/Treal(bp[i]);
			}
			global::nops+= N;
		}
		break;
	case pair(Kint,KRealArray):
		{
			X = Treal(a.IntValue());
			N = b.length();
			c.copydimsRealArray(b);
			Treal *cp=c.RealPtr(), *bp=b.RealPtr();
			for (i=0; i<N; i++) cp[i] = X/bp[i];
			global::nops+= N;
		}
		break;
	case pair(Kint,KComplexArray):
		{
			Z = Tcomplex(Treal(a.IntValue()),0.0);
			N = b.length();
			c.copydimsComplexArray(b);
			Tcomplex *cp=c.ComplexPtr(), *bp=b.ComplexPtr();
			for (i=0; i<N; i++) cp[i] = Z/bp[i];
			global::nops+= N;
		}
		break;
	case pair(Kreal,Kint):
		c = a.RealValue()/Treal(b.IntValue());
		global::nops++;
		break;
	case pair(Kreal,Kreal):
		c = a.RealValue()/b.RealValue();
		global::nops++;
		break;
	case pair(Kreal,Kcomplex):
		c = Tcomplex(a.RealValue())/b.ComplexValue();
		global::nops++;
		break;
	case pair(Kreal,KIntArray):
		{
			N = b.length();
			X = a.RealValue();
			Tint *bp=b.IntPtr();
			if (&c == &b) {
				Tobject c1;
				c1.copydimsRealArray(b);
				Treal *c1p=c1.RealPtr();
				for (i=0; i<N; i++) c1p[i] = X/bp[i];
				c = c1;
			} else {
				c.copydimsRealArray(b);	// make c a real array of the same dims as b
				Treal *cp=c.RealPtr();
				for (i=0; i<N; i++) cp[i] = X/bp[i];
			}
			global::nops+= N;
		}
		break;
	case pair(Kreal,KRealArray):
		{
			N = b.length();
			X = a.RealValue();
			c.copydimsRealArray(b);
			Treal *cp=c.RealPtr(), *bp=b.RealPtr();
			for (i=0; i<N; i++) cp[i] = X/bp[i];
			global::nops+= N;
		}
		break;
	case pair(Kreal,KComplexArray):
		{
			Z = a.RealValue();
			N = b.length();
			c.copydimsComplexArray(b);
			Tcomplex *cp=c.ComplexPtr(), *bp=b.ComplexPtr();
			for (i=0; i<N; i++) cp[i] = Z/bp[i];
			global::nops+= N;
		}
		break;
	case pair(Kcomplex,Kint):
		c = a.ComplexValue()/Tcomplex(b.IntValue());
		global::nops++;
		break;
	case pair(Kcomplex,Kreal):
		c = a.ComplexValue()/Tcomplex(b.RealValue());
		global::nops++;
		break;
	case pair(Kcomplex,Kcomplex):
		c = a.ComplexValue()/b.ComplexValue();
		global::nops++;
		break;
	case pair(Kcomplex,KIntArray):
		{
			N = b.length();
			Z = a.ComplexValue();
			Tint *bp=b.IntPtr();
			if (&c == &b) {
				Tobject c1;
				c1.copydimsComplexArray(b);
				Tcomplex *c1p=c1.ComplexPtr();
				for (i=0; i<N; i++) c1p[i] = Z/bp[i];
				c = c1;
			} else {
				c.copydimsComplexArray(b);
				Tcomplex *cp=c.ComplexPtr();
				for (i=0; i<N; i++) cp[i] = Z/bp[i];
			}
			global::nops+= N;
		}
		break;
	case pair(Kcomplex,KRealArray):
		{
			N = b.length();
			Z = a.ComplexValue();
			Treal *bp=b.RealPtr();
			if (&c == &b) {
				Tobject c1;
				c1.copydimsComplexArray(b);
				Tcomplex *c1p=c1.ComplexPtr();
				for (i=0; i<N; i++) c1p[i] = Z/bp[i];
				c = c1;
			} else {
				c.copydimsComplexArray(b);
				Tcomplex *cp=c.ComplexPtr();
				for (i=0; i<N; i++) cp[i] = Z/bp[i];
			}
			global::nops+= N;
		}
		break;
	case pair(Kcomplex,KComplexArray):
		{
			N = b.length();
			Z = a.ComplexValue();
			c.copydimsComplexArray(b);
			Tcomplex *cp=c.ComplexPtr(), *bp=b.ComplexPtr();
			for (i=0; i<N; i++) cp[i] = Z/bp[i];
			global::nops+= N;
		}
		break;
	case pair(KIntArray,Kint):
		{
			X = 1.0/Treal(b.IntValue());
			N = a.length();
			Tint *ap=a.IntPtr();
			if (&a == &c) {
				Tobject c1;
				c1.copydimsRealArray(a);
				Treal *c1p=c1.RealPtr();
				for (i=0; i<N; i++) c1p[i] = ap[i]*X;
				c = c1;
			} else {
				c.copydimsRealArray(a);
				Treal *cp=c.RealPtr();
				for (i=0; i<N; i++) cp[i] = ap[i]*X;
			}
			global::nops+= N;
		}
		break;
	case pair(KIntArray,Kreal):
		{
			N = a.length();
			X = 1.0/b.RealValue();
			Tint *ap=a.IntPtr();
			if (&c == &a) {
				Tobject c1;
				c1.copydimsRealArray(a);
				Treal *c1p=c1.RealPtr();
				for (i=0; i<N; i++) c1p[i] = ap[i]*X;
				c = c1;
			} else {
				c.copydimsRealArray(a);
				Treal *cp=c.RealPtr();
				for (i=0; i<N; i++) cp[i] = ap[i]*X;
			}
			global::nops+= N;
		}
		break;
	case pair(KIntArray,Kcomplex):
		{
			N = a.length();
			Z = 1/b.ComplexValue();
			Tint *ap=a.IntPtr();
			if (&c == &a) {
				Tobject c1;
				c1.copydimsComplexArray(a);
				Tcomplex *c1p=c1.ComplexPtr();
				for (i=0; i<N; i++) c1p[i] = ap[i]*Z;
				c = c1;
			} else {
				c.copydimsComplexArray(a);
				Tcomplex *cp=c.ComplexPtr();
				for (i=0; i<N; i++) cp[i] = ap[i]*Z;
			}
			global::nops+= N;
		}
		break;
	case pair(KIntArray,KIntArray):
		N = a.length();
		//if (N != b.length()) LengthError(OPCHAR,a,b);
		if (!a.iscompatible(b)) LengthError(OPCHAR,a,b);
		if (&c == &a || &c == &b) {
			Tobject c1;
			c1.copydimsRealArray(a);
			Treal *c1p=c1.RealPtr(); Tint *ap=a.IntPtr(), *bp=b.IntPtr();
			for (i=0; i<N; i++) c1p[i] = Treal(ap[i])/Treal(bp[i]);
			c = c1;
		} else {
			c.copydimsRealArray(a);
			Treal *cp=c.RealPtr(); Tint *ap=a.IntPtr(), *bp=b.IntPtr();
			for (i=0; i<N; i++) cp[i] = Treal(ap[i])/Treal(bp[i]);
		}
		global::nops+= N;
		break;
	case pair(KIntArray,KRealArray):
		{
			N = a.length();
			//if (N != b.length()) LengthError(OPCHAR,a,b);
			if (!a.iscompatible(b)) LengthError(OPCHAR,a,b);
			if (&a == &c) {
				Tobject c1;
				c1.copydimsRealArray(b);
				Treal *cp=c1.RealPtr(), *bp=b.RealPtr(); Tint *ap=a.IntPtr();
				for (i=0; i<N; i++) cp[i] = Treal(ap[i])/bp[i];
				c = c1;
			} else {
				c.copydimsRealArray(b);
				Treal *cp=c.RealPtr(), *bp=b.RealPtr(); Tint *ap=a.IntPtr();
				for (i=0; i<N; i++) cp[i] = Treal(ap[i])/bp[i];
			}
			global::nops+= N;
		}
		break;
	case pair(KIntArray,KComplexArray):
		{
			N = a.length();
			//if (N != b.length()) LengthError(OPCHAR,a,b);
			if (!a.iscompatible(b)) LengthError(OPCHAR,a,b);
			Tint *ap=a.IntPtr(); Tcomplex *bp=b.ComplexPtr();
			if (&a == &c) {
				Tobject c1;
				c1.copydimsComplexArray(b);
				Tcomplex *c1p=c1.ComplexPtr();
				for (i=0; i<N; i++) c1p[i] = ap[i]/bp[i];
				c = c1;
			} else {
				c.copydimsComplexArray(b);
				Tcomplex *cp=c.ComplexPtr();
				for (i=0; i<N; i++) cp[i] = ap[i]/bp[i];
			}
			global::nops+= N;
		}
		break;
	case pair(KRealArray,Kint):
		{
			X = 1.0/Treal(b.IntValue());
			N = a.length();
			c.copydimsRealArray(a);
			Treal *cp=c.RealPtr(), *ap=a.RealPtr();
			for (i=0; i<N; i++) cp[i] = ap[i]*X;
			global::nops+= N;
		}
		break;
	case pair(KRealArray,Kreal):
		{
			N = a.length();
			X = 1.0/b.RealValue();
			c.copydimsRealArray(a);
			Treal *cp=c.RealPtr(), *ap=a.RealPtr();
			for (i=0; i<N; i++) cp[i] = ap[i]*X;
			global::nops+= N;
		}
		break;
	case pair(KRealArray,Kcomplex):
		{
			N = a.length();
			Treal *ap=a.RealPtr();
			Z = 1/b.ComplexValue();
			if (&c == &a) {
				Tobject c1;
				c1.copydimsComplexArray(a);
				Tcomplex *c1p=c1.ComplexPtr();
				for (i=0; i<N; i++) c1p[i] = ap[i]*Z;
				c = c1;
			} else {
				c.copydimsComplexArray(a);
				Tcomplex *cp=c.ComplexPtr();
				for (i=0; i<N; i++) cp[i] = ap[i]*Z;
			}
			global::nops+= N;
		}
		break;
	case pair(KRealArray,KIntArray):
		{
			N = a.length();
			//if (N != b.length()) LengthError(OPCHAR,a,b);
			if (!a.iscompatible(b)) LengthError(OPCHAR,a,b);
			Tint *bp=b.IntPtr();
			Treal *ap=a.RealPtr();
			if (&b == &c) {
				Tobject c1;
				c1.copydimsRealArray(a);
				Treal *c1p=c1.RealPtr();
				for (i=0; i<N; i++) c1p[i] = ap[i]/Treal(bp[i]);
				c = c1;
			} else {
				c.copydimsRealArray(a);
				Treal *cp=c.RealPtr();
				for (i=0; i<N; i++) cp[i] = ap[i]/Treal(bp[i]);
			}
			global::nops+= N;
		}
		break;
	case pair(KRealArray,KRealArray):
		{
			c.copydimsRealArray(a);
			N = a.length();
			//if (N != b.length()) LengthError(OPCHAR,a,b);
			if (!a.iscompatible(b)) LengthError(OPCHAR,a,b);
			Treal *cp=c.RealPtr(), *ap=a.RealPtr(), *bp=b.RealPtr();
			for (i=0; i<N; i++) cp[i] = ap[i]/bp[i];
			global::nops+= N;
		}
		break;
	case pair(KRealArray,KComplexArray):
		{
			N = a.length();
			//if (N != b.length()) LengthError(OPCHAR,a,b);
			if (!a.iscompatible(b)) LengthError(OPCHAR,a,b);
			Treal *ap=a.RealPtr(); Tcomplex *bp=b.ComplexPtr();
			if (&a == &c) {
				Tobject c1;
				c1.copydimsComplexArray(b);
				Tcomplex *c1p=c1.ComplexPtr();
				for (i=0; i<N; i++) c1p[i] = ap[i]/bp[i];
			} else {
				c.copydimsComplexArray(b);
				Tcomplex *cp=c.ComplexPtr();
				for (i=0; i<N; i++) cp[i] = ap[i]/bp[i];
			}
			global::nops+= N;
		}
		break;
	case pair(KComplexArray,Kint):
		{
			N = a.length();
			Z = 1/Tcomplex(b.IntValue());
			c.copydimsComplexArray(a);
			Tcomplex *cp=c.ComplexPtr(), *ap=a.ComplexPtr();
			for (i=0; i<N; i++) cp[i] = ap[i]*Z;
			global::nops+= N;
		}
		break;
	case pair(KComplexArray,Kreal):
		{
			N = a.length();
			X = 1.0/b.RealValue();
			c.copydimsComplexArray(a);
			Tcomplex *cp=c.ComplexPtr(), *ap=a.ComplexPtr();
			for (i=0; i<N; i++) cp[i] = ap[i]*X;
			global::nops+= N;
		}
		break;
	case pair(KComplexArray,Kcomplex):
		{
			N = a.length();
			Z = 1/b.ComplexValue();
			c.copydimsComplexArray(a);
			Tcomplex *cp=c.ComplexPtr(), *ap=a.ComplexPtr();
			for (i=0; i<N; i++) cp[i] = ap[i]*Z;
			global::nops+= N;
		}
		break;
	case pair(KComplexArray,KIntArray):
		{
			N = a.length();
			//if (N != b.length()) LengthError(OPCHAR,a,b);
			if (!a.iscompatible(b)) LengthError(OPCHAR,a,b);
			Tcomplex *ap=a.ComplexPtr(); Tint *bp=b.IntPtr();
			if (&b == &c) {
				Tobject c1;
				c1.copydimsComplexArray(a);
				Tcomplex *c1p=c1.ComplexPtr();
				for (i=0; i<N; i++) c1p[i] = ap[i]/bp[i];
				c = c1;
			} else {
				c.copydimsComplexArray(a);
				Tcomplex *cp=c.ComplexPtr();
				for (i=0; i<N; i++) cp[i] = ap[i]/bp[i];
			}
			global::nops+= N;
		}
		break;
	case pair(KComplexArray,KRealArray):
		{
			N = a.length();
			//if (N != b.length()) LengthError(OPCHAR,a,b);
			if (!a.iscompatible(b)) LengthError(OPCHAR,a,b);
			Tcomplex *ap=a.ComplexPtr(); Treal *bp=b.RealPtr();
			if (&b == &c) {
				Tobject c1;
				c1.copydimsComplexArray(a);
				Tcomplex *c1p=c1.ComplexPtr();
				for (i=0; i<N; i++) c1p[i] = ap[i]/bp[i];
				c = c1;
			} else {
				c.copydimsComplexArray(a);
				Tcomplex *cp=c.ComplexPtr();
				for (i=0; i<N; i++) cp[i] = ap[i]/bp[i];
			}
			global::nops+= N;
		}
		break;
	case pair(KComplexArray,KComplexArray):
		{
			c.copydimsComplexArray(b);
			N = a.length();
			//if (N != b.length()) LengthError(OPCHAR,a,b);
			if (!a.iscompatible(b)) LengthError(OPCHAR,a,b);
			Tcomplex *cp=c.ComplexPtr(), *bp=b.ComplexPtr(), *ap=a.ComplexPtr();
			for (i=0; i<N; i++) cp[i] = ap[i]/bp[i];
			global::nops+= N;
		}
		break;
	default:
		BinError(OPCHAR,a,b);
		break;
	}
} // Div

#if !defined(HPUX) || defined(__GNUC__)
  inline Treal pow(Treal x, Tint y) {return pow(x,Treal(y));}
#endif
inline Tint pow(Tint x, Tint y) {return round(pow(Treal(x),y));}
inline Treal pow(Tint x, Treal y) {return pow(Treal(x),y);}

#define OP(a,b) pow(a,b)
#define OPCHAR "Pow"
#define OPFNAME Pow
#include "templ/tbinop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

inline Tint modulus(Tint x, Tint y) {return (y==0) ? 0 : x % y;}
inline Treal modulus(Treal x, Tint y) {return (y==0) ? 0 : fmod(x,y);}
inline Treal modulus(Tint x, Treal y) {return (y==0) ? 0 : fmod(x,y);}
inline Treal modulus(Treal x, Treal y) {return (y==0) ? 0 : fmod(x,y);}
static Tcomplex modulus(Tcomplex x, Tint y) {return Tcomplex((y==0) ? 0 : fmod(real(x),y),0);}
static Tcomplex modulus(Tcomplex x, Treal y) {return Tcomplex((y==0) ? 0 : fmod(real(x),y),0);}
static Tcomplex modulus(Tcomplex x, Tcomplex y) {
	return Tcomplex((real(y)==0) ? 0 : fmod(real(x),real(y)), (imag(y)==0) ? 0 : fmod(imag(x),imag(y)));
}
inline Tcomplex modulus(Tint x, Tcomplex y) {return modulus(Tcomplex(x),y);}
inline Tcomplex modulus(Treal x, Tcomplex y) {return modulus(Tcomplex(x),y);}

#define OP(a,b) modulus(a,b)
#define OPCHAR "mod"
#define OPFNAME Mod
#include "templ/tbinop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

// Comparison operators >,<,>=,<=

#define OP(a,b) ((a)>(b))
#define OPCHAR ">"
#define OPFNAME Gt
#include "templ/tcmpop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a,b) ((a)<(b))
#define OPCHAR "<"
#define OPFNAME Lt
#include "templ/tcmpop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a,b) ((a)>=(b))
#define OPCHAR ">="
#define OPFNAME Ge
#include "templ/tcmpop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a,b) ((a)<=(b))
#define OPCHAR "<="
#define OPFNAME Le
#include "templ/tcmpop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

// Equality operators == and !=

#define OP(a,b) ((a)==(b))
#define OPCHAR "=="
#define OPFNAME Eq
#include "templ/teqop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a,b) ((a)!=(b))
#define OPCHAR "!="
#define OPFNAME Ne
#include "templ/teqop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

// Logical operators And, Or

#define OP(a,b) ((a)&&(b))
#define OPCHAR "&&"
#define OPFNAME And
#include "templ/tbinintop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a,b) ((a)||(b))
#define OPCHAR "||"
#define OPFNAME Or
#include "templ/tbinintop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

// Comparison tests >,<,>=,<= (these return int code, not assign to objects)

#define OP(a,b) ((a)>(b))
#define OPCHAR ">"
#define OPFNAME TestGt
#include "templ/ttestcmpop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a,b) ((a)<(b))
#define OPCHAR "<"
#define OPFNAME TestLt
#include "templ/ttestcmpop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a,b) ((a)>=(b))
#define OPCHAR ">="
#define OPFNAME TestGe
#include "templ/ttestcmpop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a,b) ((a)<=(b))
#define OPCHAR "<="
#define OPFNAME TestLe
#include "templ/ttestcmpop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

// Binary Min/Max operators

#define OP(a,b) min(a,b)
#define OPCHAR "Min"
#define OPFNAME Min
#include "templ/tminmaxop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a,b) max(a,b)
#define OPCHAR "Max"
#define OPFNAME Max
#include "templ/tminmaxop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

// Unary Min/Max operators

void Min(Tobject& c, const Tobject& a)
/* Min(c,a) returns the minimum element of a in c.
   The object a must not be complex. If a is scalar,
   it is returned as such. If it is an array, the
   minimum element (scalar) is returned. */
{
    switch (a.kind()) {
    case Kint:
        c = a.IntValue();
        break;
    case Kreal:
        c = a.RealValue();
        break;
    case KIntArray:
        {
            Tint *const p = a.IntPtr();
            Tint m = p[0];
            for (int i=1; i<a.length(); i++) m = min(m,p[i]);
            c = m;
        }
        break;
    case KRealArray:
        {
            Treal *const p = a.RealPtr();
            Treal m = p[0];
            for (int i=1; i<a.length(); i++) m = min(m,p[i]);
            c = m;
        }
        break;
	default:
		UnaryError("Min",a);
		break;
    }
}

void Max(Tobject& c, const Tobject& a)
/* Max(c,a) returns the maximum element of a in c.
   The object a must not be complex. If a is scalar,
   it is returned as such. If it is an array, the
   maximum element (scalar) is returned. */
{
    switch (a.kind()) {
    case Kint:
        c = a.IntValue();
        break;
    case Kreal:
        c = a.RealValue();
        break;
    case KIntArray:
        {
            Tint *const p = a.IntPtr();
            Tint m = p[0];
            for (int i=1; i<a.length(); i++) m = max(m,p[i]);
            c = m;
        }
        break;
    case KRealArray:
        {
            Treal *const p = a.RealPtr();
            Treal m = p[0];
            for (int i=1; i<a.length(); i++) m = max(m,p[i]);
            c = m;
        }
        break;
	default:
		UnaryError("Max",a);
		break;
    }
}


// The Not operator does not fit any of the templates

void Not(Tobject& c, const Tobject& a) {
	int i;
	switch (a.kind()) {
	case Kint:
		c = !(a.IntValue());
		global::nops++;
		break;
	case KIntArray:
		{
			int N = a.length();
			c.copydimsIntArray(a);
			Tint *cp=c.IntPtr(), *ap=a.IntPtr();;
			for (i=0; i<N; i++) cp[i] = !(ap[i]);
			global::nops+= N;
		}
		break;
	default:
		UnaryError("!",a);
		break;
	}
}
		

// Unary operators -, Inc, Dec

#define OP(a) (-(a))
#define OPCHAR "-"
#define OPFNAME Neg
#include "templ/tunop.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

inline void Increment(Tint& i) {i++;}
inline void Increment(Treal& x) {x+=1;}
inline void Increment(Tcomplex& z) {z+=1;}
inline void Decrement(Tint& i) {i--;}
inline void Decrement(Treal& x) {x-=1;}
inline void Decrement(Tcomplex& z) {z-=1;}

#define OP(a) Increment(a)
#define OPCHAR "++"
#define OPFNAME Inc
#include "templ/tunop1.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

#define OP(a) Decrement(a)
#define OPCHAR "--"
#define OPFNAME Dec
#include "templ/tunop1.C"
#undef OPFNAME
#undef OPCHAR
#undef OP

// The Abs operator doesn't fit to the standard template schemes
// since abs(complex) -> real

void Abs(Tobject& c, const Tobject& a) {
	int i;
	switch (a.kind()) {
	case Kint:
		c = abs(a.IntValue());
		c.ClearCharFlag();
		global::nops++;
		break;
	case Kreal:
		c = fabs(a.RealValue());
		global::nops++;
		break;
	case Kcomplex:
		c = abs(a.ComplexValue());
		global::nops++;
		break;
	case KIntArray:
		{
			c.copydimsIntArray(a);
			int N = a.length();
			Tint *cp=c.IntPtr(), *ap=a.IntPtr();
			for (i=0; i<N; i++) cp[i] = abs(ap[i]);
			c.ClearStringFlag();
			global::nops+= N;
		}
		break;
	case KRealArray:
		{
			c.copydimsRealArray(a);
			int N = a.length();
			Treal *cp=c.RealPtr(), *ap=a.RealPtr();
			for (i=0; i<N; i++) cp[i] = fabs(ap[i]);
			global::nops+= N;
		}
		break;
	case KComplexArray:
		{
			int N = a.length();
			if (&c == &a) {
				Tobject c1;
				c1.copydimsRealArray(a);
				Treal *c1p=c1.RealPtr(); Tcomplex *ap=a.ComplexPtr();
				for (i=0; i<N; i++) c1p[i] = abs(ap[i]);
				c = c1;
			} else {
				c.copydimsRealArray(a);
				Treal *cp=c.RealPtr(); Tcomplex *ap=a.ComplexPtr();
				for (i=0; i<N; i++) cp[i] = abs(ap[i]);
			}
			global::nops+= N;
		}
		break;
	case KObjectArray:
	case Kfunction:
	case KCfunction:
	case KIntrinsicFunction:
	case Kvoid:
	case Kundef:
		UnaryError("Abs",a);
		break;
	}
}
		

// A general one-argument transcendental function FN, which is defined
// at all complex values, and is real-valued on the real axis.
// Transcen(FNpack,c,a) performes c=FN(a) for arbitrary numeric object a.
// FNpack is a struct containing the name ("FN"), plus pointers to 4 evaluators:
// real scalar FN, complex scalar FN, real vector FN, complex vector FN.

void Transcen(const TTranscenFunctionPack& funcs, Tobject& c, const Tobject& a) {
	int i,N;
	switch (a.kind()) {
	case Kint:
		c = (*funcs.f)(Treal(a.IntValue()));
		global::nops++;
		break;
	case Kreal:
		c = (*funcs.f)(a.RealValue());
		global::nops++;
		break;
	case Kcomplex:
		c = (*funcs.cf)(a.ComplexValue());
		global::nops++;
		break;
	case KIntArray:
		N = a.length();
		if (&c==&a) {
			Tobject c1;
			c1.copydimsRealArray(a);
			Treal *c1p=c1.RealPtr(); Tint *ap=a.IntPtr();
			for (i=0; i<N; i++) c1p[i] = Treal(ap[i]);
			(*funcs.vf)(c1p,N);
			c = c1;
		} else {
			c.copydimsRealArray(a);
			Treal *cp=c.RealPtr(); Tint *ap=a.IntPtr();
			for (i=0; i<N; i++) cp[i] = Treal(ap[i]);
			(*funcs.vf)(cp,N);
		}
		global::nops+= N;
		break;
	case KRealArray:
		{
			N = a.length();
			if (&c!=&a) c.copydimsRealArray(a);
			Treal *cp=c.RealPtr(), *ap=a.RealPtr();
			for (i=0; i<N; i++) cp[i] = ap[i];
			(*funcs.vf)(cp,N);
			global::nops+= N;
		}
		break;
	case KComplexArray:
		{
			N = a.length();
			if (&c!=&a) c.copydimsComplexArray(a);
			Tcomplex *cp=c.ComplexPtr(), *ap=a.ComplexPtr();
			for (i=0; i<N; i++) cp[i] = ap[i];
			(*funcs.cvf)(cp,N);
			global::nops+= N;
		}
		break;
	case KObjectArray:
	case Kfunction:
	case KCfunction:
	case KIntrinsicFunction:
	case Kvoid:
	case Kundef:
		err << "Transcendental function '" << funcs.name << "' applied to " << Tshort(a) << "\n";
		error();
		break;
	}
}


